"""Top-level package for GISMO."""
from importlib.metadata import metadata

infos = metadata(__name__)
__author__ = infos['Author']
__email__ = infos['Author-Email']
__version__ = infos['Version']


from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.diteration import DIteration
from gismo.gismo import Gismo, XGismo



from spacy.lang.en import English

from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.gismo import Gismo


class Sentencizer:
    """
    The Sentencizer class allows to refine a document-level gismo into a sentence-level gismo.
    A simple sentence extraction is proposed.
    For more complex usages, the class can provide a full :py:class:`~gismo.gismo.Gismo` instance that operates at
    sentence-level.

    Parameters
    ----------
    gismo: Gismo
        Document-level Gismo.

    Examples
    ---------
    We use the C50 Reuters dataset (5000 news paragraphs).

    >>> from gismo.datasets.reuters import get_reuters_news
    >>> corpus = Corpus(get_reuters_news(), to_text=lambda e: e['content'])
    >>> embedding = Embedding()
    >>> embedding.fit_transform(corpus)
    >>> gismo = Gismo(corpus, embedding)
    >>> sentencer = Sentencizer(gismo)

    First Example: run explicitly the query *Orange* at document-level,
    extract 4 covering sentences with narrow bfs.

    >>> success = gismo.rank("Orange")
    >>> sentencer.get_sentences(s=4, wide=False) # doctest: +NORMALIZE_WHITESPACE
    ['Snook says the all important average retained revenue per Orange subscriber will
    rise from around 442 pounds per year, partly because dominant telecoms player British
    Telecommunications last month raised the price of a call to Orange phones from its fixed lines.',
    'Analysts said that Orange shares had good upside potential after a rollercoaster ride in their
    short time on the market.',
    'Orange, which was floated last March at 205 pence per share, initially saw its stock slump to
    157.5 pence before recovering over the last few months to trade at 218 on Tuesday, a rise of four
    pence on the day.',
    'One-2-One and Orange ORA.L, which offer only digital services, are due to release their
    connection figures next week.']

    Second example: extract *Ericsson*-related sentences

    >>> sentencer.get_sentences(query="Ericsson") # doctest: +NORMALIZE_WHITESPACE
    ['These latest wins follow a recent $350 million contract win with Telefon AB L.M. Ericsson,
    bolstering its already strong activity in the contract manufacturing of telecommuncation and
    data communciation products, he said.',
    'The restraints are few in areas such as consumer products, while in sectors such as banking,
    distribution and insurance, foreign firms are kept on a very tight leash.',
    "The company also said it had told analysts in a briefing Tuesday of new contract wins with
    Ascend Communications Inc, Harris Corp's Communications unit and Philips Electronics NV.",
    'Pocket is the first from the high-priced 1996 auction known to have filed for bankruptcy
    protection.',
    'With Ascend in particular, he said the company would be manufacturing the company\\'s
    mainstream MAX TNT remote access network equipment. "']

    Third example: extract *Communications*-related sentences from a string.

    >>> txt = gismo.corpus[4517]['content']
    >>> sentencer.get_sentences(query="communications", txt=txt) # doctest: +NORMALIZE_WHITESPACE
    ["Privately-held Pocket's big creditors include a group of Asian entrepreneurs and
    communications-equipment makers Siemens AG of Germany and L.M. Ericsson of Sweden.",
    "2 bidder at the government's high-flying wireless phone auction last year has filed for
    bankruptcy protection from its creditors, underscoring the problems besetting the
    auction's winners.",
    "The Federal Communications Commission on Monday gave PCS companies from last year's
    auction some breathing space when it suspended indefinitely a March 31 deadline for
    them to make payments to the agency for their licenses."]
    """

    def __init__(self, gismo):
        self.parser = English()
        self.parser.add_pipe('sentencizer')
        self.doc_gismo = gismo
        self.sent_corpus = None
        self.sent_gismo = None

    def splitter(self, txt):
        """
        Transform input content into a corpus of sentences stored into the :py:attr:`sent_corpus` attribute.

        Parameters
        ----------
        txt: str or list
            Text or list of documents to split in sentences. For the latter, documents are assumed
            to be provided as `(content, id)` pairs, where `content` is the actual text and `id` a
            reference of the document.

        Returns
        -------
        Sentencizer
        """
        if type(txt) is str:
            source = [str(sent).strip() for sent in self.parser(txt).sents if len(sent) > 10]
            self.sent_corpus = Corpus(source, to_text=lambda x: x)
        else:
            source = [{'source': source[1],
                       'content': str(sent).strip()} for source in txt
                      for sent in self.parser(source[0]).sents if len(sent) > 10]
            self.sent_corpus = Corpus(source, to_text=lambda x: x['content'])
        return self

    def make_sent_gismo(self, query=None, txt=None, k=None, **kwargs):
        """
        Construct a sentence-level Gismo stored in the :py:attr:`sent_gismo` attribute.

        Parameters
        ----------
        query: str (optional)
            Query to run on the document-level Gismo.
        txt: str (optional)
            Text to use for sentence extraction.
            If not set, the sentences will be extracted from the top-documents.
        k: int (optional)
            Number of top-documents used for the built.
            If not set, the :py:func:`~gismo.common.auto_k` heuristic will be used.
        kwargs: dict
            Custom default runtime parameters to pass to the sentence-level Gismo.
            You just need to specify the parameters that differ from :obj:`~gismo.parameters.DEFAULT_PARAMETERS`.
            Note that distortion will be automatically de-activated. If you really want it, manually change the value
            of ``self.sent_gismo.parameters.distortion`` afterwards.


        Returns
        -------
        Sentencizer
        """
        if txt is None:
            if query is not None:
                self.doc_gismo.rank(query)
            txt = [(self.doc_gismo.corpus.to_text(self.doc_gismo.corpus[i]), i)
                   for i in self.doc_gismo.get_documents_by_rank(k, post=False)]
        self.splitter(txt)
        local_embedding = Embedding()
        local_embedding.fit_ext(self.doc_gismo.embedding)
        local_embedding.transform(self.sent_corpus)
        self.sent_gismo = Gismo(self.sent_corpus, local_embedding, **kwargs)
        self.sent_gismo.parameters.distortion = 0.0
        self.sent_gismo.post_documents_item = lambda g, i: g.corpus.to_text(g.corpus[i])
        return self

    def get_sentences(self, query=None, txt=None, k=None, s=None,
                      resolution=.7, stretch=2.0, wide=True, post=True):
        """
        All-in-one method to extract covering sentences from the corpus.
        Computes sentence-level corpus, sentence-level gismo,
        and calls :py:meth:`~gismo.gismo.Gismo.get_documents_by_coverage`.

        Parameters
        ----------
        query: str (optional)
            Query to run on the document-level Gismo
        txt: str (optional)
            Text to use for sentence extraction.
            If not set, the sentences will be extracted from the top-documents.
        k: int (optional)
            Number of top-documents used for the built.
            If not set, the :py:func:`~gismo.common.auto_k` heuristic of the document-level Gismo will be used.
        s: int (optional)
            Number of sentences to return.
            If not set, the :py:func:`~gismo.common.auto_k` heuristic of the sentence-level Gismo will be used.
        resolution: float (optional)
            Tree resolution passed to the :py:meth:`~gismo.gismo.Gismo.get_documents_by_coverage` method.
        stretch: float >= 1 (optional)
            Stretch factor passed to the :py:meth:`~gismo.gismo.Gismo.get_documents_by_coverage` method.
        wide: bool (optional)
            bfs wideness passed to the :py:meth:`~gismo.gismo.Gismo.get_documents_by_coverage` method.
        post: bool (optional)
            Use of post-proccessing passed to the :py:meth:`~gismo.gismo.Gismo.get_documents_by_coverage` method.

        Returns
        -------
        list
        """

        self.make_sent_gismo(query=query, txt=txt, k=k)
        if query is None:
            query = self.doc_gismo.embedding._query
        self.sent_gismo.rank(query)
        return self.sent_gismo.get_documents_by_coverage(k=s, resolution=resolution,
                                                         stretch=stretch, wide=wide, post=post)



import requests
import urllib
import pkgutil
import gzip
import json

from bs4 import BeautifulSoup as bs


def get_acm_from_package():
    data = pkgutil.get_data(__package__, 'acm.json.gz')
    return json.loads(gzip.decompress(data))


def get_acm():
    """
    Returns
    -------
    acm: list of dicts
        Each dict is an ACM domain. It contains category name,
        query (concatenation of names from domain and subdomains),
        size (number of subdomains including itself), and children (list of domain dicts).

    Examples
    --------

    >>> acm = get_acm()
    >>> subdomain = acm[4]['children'][2]['children'][1]
    >>> subdomain['name']
    'Software development process management'
    >>> subdomain['size']
    10
    >>> subdomain['query']
    'Software development process management, Software development methods, Rapid application development, Agile software development, Capability Maturity Model, Waterfall model, Spiral model, V-model, Design patterns, Risk management'
    >>> len(acm)
    13
    """
    return get_acm_from_package()


def flatten_acm(acm, min_size=5, max_depth=100, exclude=None, depth=0):
    """
    Select subdomains of an acm tree and return them as a list.

    Parameters
    ----------
    acm: list of dicts
        acm tree from get_acm.
    min_size: int
        size threshold to select a domain (avoids small domains)
    max_depth: int
        depth threshold to select a domain (avoids deep domains)
    exclude: list
        list of domains to exclude from the results

    Returns
    -------
    list
        A flat list of domains described by name and query.

    Example
    -------
    >>> acm = flatten_acm(get_acm())
    >>> acm[111]['name']
    'Graph theory'
    """
    if exclude is None:
        exclude = set()
    result = [{'name': t['name'], 'query': t['query']} for t in acm if t['size'] > min_size]
    for t in acm:
        if len(t['children']) > 0 and depth < max_depth:
            result += flatten_acm(t['children'], min_size=min_size, max_depth=max_depth, depth=depth + 1)
    if depth == 0:
        result = [{'name': key, 'query': value}
                  for key, value in {t['name']: t['query'] for t in result}.items()
                  if key not in exclude]
    return result



from numba import njit
import numpy as np

from gismo.parameters import ALPHA, N_ITER, OFFSET, MEMORY


# diffusion: starting point is on the feature (words) space
# provides ranking on features (X) and documents (Y)
@njit
def jit_diffusion(x_pointers, x_indices, x_data,
                  y_pointers, y_indices, y_data,
                  z_indices, z_data,
                  x_relevance, y_relevance,
                  alpha, n_iter, offset: float,
                  x_fluid, y_fluid):
    """
    Core diffusion engine written to be compatible with `Numba <https://numba.pydata.org/>`_.
    This is where the `DIteration <https://arxiv.org/pdf/1501.06350.pdf>`_
    algorithm is applied inline.

    Parameters
    ----------
    x_pointers: :class:`~numpy.ndarray`
        Pointers of the :class:`~scipy.sparse.csr_matrix` embedding of documents.
    x_indices: :class:`~numpy.ndarray`
        Indices of the :class:`~scipy.sparse.csr_matrix` embedding of documents.
    x_data: :class:`~numpy.ndarray`
        Data of the :class:`~scipy.sparse.csr_matrix` embedding of documents.
    y_pointers: :class:`~numpy.ndarray`
        Pointers of the :class:`~scipy.sparse.csr_matrix` embedding of features.
    y_indices: :class:`~numpy.ndarray`
        Indices of the :class:`~scipy.sparse.csr_matrix` embedding of features.
    y_data: :class:`~numpy.ndarray`
        Data of the :class:`~scipy.sparse.csr_matrix` embedding of features.
    z_indices: :class:`~numpy.ndarray`
        Indices of the :class:`~scipy.sparse.csr_matrix` embedding of the query projection.
    z_data: :class:`~numpy.ndarray`
        Data of the :class:`~scipy.sparse.csr_matrix` embedding of the query_projection.
    x_relevance: :class:`~numpy.ndarray`
        Placeholder for relevance of documents.
    y_relevance: :class:`~numpy.ndarray`
        Placeholder for relevance of features.
    alpha: float in range [0.0, 1.0]
        Damping factor. Controls the trade-off between closeness and centrality.
    n_iter: int
        Number of round-trip diffusions to perform. Higher value means better precision
        but longer execution time.
    offset: float in range [0.0, 1.0]
        Controls how much of the initial fluid should be deduced form the relevance.
    x_fluid: :class:`~numpy.ndarray`
        Placeholder for fluid on the side of documents.
    y_fluid: :class:`~numpy.ndarray`
        Placeholder for fluid on the side of features.
    """
    n = len(x_pointers) - 1
    m = len(y_pointers) - 1

    # Reset fluids
    x_fluid[:] = 0
    y_fluid[:] = 0
    for ind, data in zip(z_indices, z_data):
        y_relevance[ind] -= data * offset  # First round penalty
        y_fluid[ind] = data

    # Core diffusion
    for turn in range(n_iter):
        for j in range(m):
            f = y_fluid[j]
            y_fluid[j] = 0.0
            if f > 0:
                y_relevance[j] += f
                x_fluid[y_indices[y_pointers[j]:y_pointers[j + 1]]] += f * alpha * y_data[
                                                                                   y_pointers[j]:y_pointers[j + 1]]
        for i in range(n):
            f = x_fluid[i]
            x_fluid[i] = 0.0
            if f > 0:
                x_relevance[i] += f
                y_fluid[x_indices[x_pointers[i]:x_pointers[i + 1]]] += f * alpha * x_data[
                                                                                   x_pointers[i]:x_pointers[i + 1]]

    # Don't waste the last drop of fluid, it's free!
    for i in range(m):
        y_relevance[i] += y_fluid[i]


class DIteration:
    """
    This class is in charge of performing the
    `DIteration <https://arxiv.org/pdf/1501.06350.pdf>`_
    algorithm.

    Parameters
    ----------
    n: int
        Number of documents.
    m: int
        Number of features.

    Attributes
    ----------
    x_relevance: :class:`~numpy.ndarray`
        Relevance of documents.
    y_relevance: :class:`~numpy.ndarray`
        Relevance of features.
    x_order: :class:`~numpy.ndarray`
        Indices of documents sorted by relevance.
    y_order: :class:`~numpy.ndarray`
        Indices of features sorted by relevance.
    """

    def __init__(self, n, m):
        self.x_relevance = np.zeros(n)
        self.y_relevance = np.zeros(m)
        self.x_order = None
        self.y_order = None
        self._x_fluid = np.zeros(n)
        self._y_fluid = np.zeros(m)

    def __call__(self, x, y, z,
                 alpha=ALPHA, n_iter=N_ITER, offset: float = OFFSET, memory=MEMORY):
        """
        Performs DIteration algorithm and populate relevance / order vectors.

        Parameters
        ----------
        x: :class:`~scipy.sparse.csr_matrix`
            Embedding of documents in feature space.
        y: :class:`~scipy.sparse.csr_matrix`
            Embedding of features in document space.
        z:  class:`~scipy.sparse.csr_matrix`
            Embedding of query in feature space.
        alpha: float in range [0.0, 1.0]
            Damping factor. Controls the trade-off between closeness and centrality.
        n_iter: int
            Number of round-trip diffusions to perform. Higher value means better precision
            but longer execution time.
        offset: float in range [0.0, 1.0]
            Controls how much of the initial fluid should be deduced form the relevance.
        memory: float in range [0.0, 1.0]
            Controls how much of previous computation is kept
            when performing a new diffusion.
            """
        self.x_relevance[:] *= memory
        self.y_relevance[:] *= memory
        jit_diffusion(x_pointers=x.indptr, x_indices=x.indices, x_data=x.data,
                      y_pointers=y.indptr, y_indices=y.indices, y_data=y.data,
                      z_indices=z.indices, z_data=z.data,
                      x_relevance=self.x_relevance, y_relevance=self.y_relevance,
                      alpha=alpha, n_iter=n_iter, offset=offset,
                      x_fluid=self._x_fluid, y_fluid=self._y_fluid)
        self.x_order = np.argsort(-self.x_relevance)
        self.y_order = np.argsort(-self.y_relevance)



import io
from pathlib import Path
import requests
import zlib
import json
import dill as pickle
import numpy as np
import gzip
from lxml import etree


URL = "https://dblp.uni-trier.de/xml/dblp.xml.gz"
"""
URL of the full DBLP database.
"""

DTD_URL = "https://dblp.uni-trier.de/xml/dblp.dtd"
"""
URL of the dtd file (required to correctly parse non-ASCII characters).
"""

DEFAULT_FIELDS = {'type', 'title', 'authors', 'venue', 'year'}
"""
Default fields to extract.
"""

LIST_TYPE_FIELDS = {'urls', 'authors'}
"""
DBLP fields with possibly multiple entries.
"""

FIELD_REDIRECTION = {'journal': 'venue',
                     'booktitle': 'venue',
                     'author': 'authors',
                     'ee': 'urls'
                     }


def fast_iter(context, func, d=2, **kwargs):
    """
    Applies ``func`` to all xml elements of depth 1 of the xml parser ``context``. `
    ``**kwargs`` are passed to ``func``.

    Modified version of a modified version of Liza Daly's fast_iter
    Inspired by
    https://stackoverflow.com/questions/4695826/efficient-way-to-iterate-through-xml-elements

    Parameters
    ----------
    context: XMLparser
        A parser obtained from etree.iterparse
    func: function
        How to process the elements
    d: int, optional
        Depth to process elements.
    """
    depth = 0
    for event, elem in context:
        if event == 'start':
            depth += 1
        if event == 'end':
            depth -= 1
            if depth < d:
                func(elem, **kwargs)
                # It's safe to call clear() here because no descendants will be
                # accessed
                elem.clear()
                # Also eliminate now-empty references from the root node to elem
                for ancestor in elem.xpath('ancestor-or-self::*'):
                    while ancestor.getprevious() is not None:
                        del ancestor.getparent()[0]
    del context


def xml_element_to_dict(elt, fields):
    """
    Converts the xml element ``elt`` into a dict if it is a paper.

    Parameters
    ----------
    elt: Any
        a XML element.
    fields: set
        Set of entries to retrieve.

    Returns
    -------
    dict or None
        Article dictionary if element contains the attributes of an article, None otherwise.
    """
    children = elt.getchildren()
    if not children:
        return None
    dic = {"type": elt.tag} if "type" in fields else dict()
    for c in children:
        value = c.text
        key = c.tag
        key = FIELD_REDIRECTION.get(key, key)
        if key not in fields or not isinstance(value, str):
            continue
        if key in LIST_TYPE_FIELDS:
            dic.setdefault(key,[]).append(value)
        else:
            dic[key] = value
    if not dic.get('authors') or not all(key in dic for key in ['year', 'title', 'venue']):
        return None
    return dic


def element_to_source(elt, source, fields):
    """
    Test if elt is an article, converts it to dictionary and appends to source

    Parameters
    ----------
    elt: Any
        a XML element.
    source: list
        the source in construction.
    fields: set
        Set of fields to retrieve.
    """
    dic = xml_element_to_dict(elt, fields)
    if dic is not None:
        source.append(dic)


def url2source(url, fields=None):
    """
    Directly transform URL of a dblp xml into a list of dictionnary.
    Only use for datasets that fit into memory (e.g. articles from one author).
    If the dataset does not fit, consider using the Dblp class instead.

    Parameters
    ----------
    url: str
        the URL to fetch.
    fields: set
        Set of DBLP fields to capture.

    Returns
    -------
    source: list of dict
        Articles retrieved from the URL

    Example
    -------
    >>> source = url2source("https://dblp.org/pers/xx/t/Tixeuil:S=eacute=bastien.xml", fields={'authors', 'title', 'year', 'venue', 'urls'})
    >>> art = [s for s in source if s['title']=="Distributed Computing with Mobile Robots: An Introductory Survey."][0]
    >>> art['authors']
    ['Maria Potop-Butucaru', 'Michel Raynal', 'Sébastien Tixeuil']
    >>> art['urls']
    ['https://doi.org/10.1109/NBiS.2011.55', 'https://doi.ieeecomputersociety.org/10.1109/NBiS.2011.55']
    """
    if fields is None:
        fields = DEFAULT_FIELDS
    r = requests.get(url)
    source = []
    with io.BytesIO(r.content) as f:
        context = etree.iterparse(f, events=('start', 'end',))
        fast_iter(context, element_to_source, d=3, source=source, fields=fields)
    return source


def element_to_filesource(elt, data_handler, index, fields):
    """
    * Converts the xml element ``elt`` into a dict if it is an article.
    * Compress and write the dict in ``data_handler``
    * Append file position in ``data_handler`` to ``index``.

    Parameters
    ----------
    elt: Any
        a XML element.
    data_handler: file_descriptor
        Where the compressed data will be stored. Must be writable.
    index:
        a list that contains the initial position of the data_handler for all previously processed elements.
    fields: set
        Set of fields to retrieve.

    Returns
    -------
    bool
        Always return True for compatibility with the xml parser.
    """
    dic = xml_element_to_dict(elt=elt, fields=fields)
    if dic is None:
        return True
    data_handler.write(zlib.compress(json.dumps(dic).encode('utf8')))
    index.append(data_handler.tell())
    return True


class Dblp:
    """
    The DBLP class can download DBLP database and produce source files compatible with the :class:`~gismo.filesource.FileSource` class.

    Parameters
    ----------
    dblp_url: str, optional
        Alternative URL for the dblp.xml.gz file
    filename: str
        Stem of the files (suffixes will be appened)
    path: str or path, optional
            Destination of the files
    """
    def __init__(self, dblp_url=URL, filename="dblp",
                 path="."):
        self.dblp_url = dblp_url
        self.path = Path(path)
        self.dblp_xml = self.path / Path(f"{filename}.xml.gz")
        self.dblp_data = self.path / Path(f"{filename}.data")
        self.dblp_index = self.path / Path(f"{filename}.index")
        self.xml_handler = None
        self.json_handler = None
        self._index = None

    def download(self):
        r = requests.get(self.dblp_url, stream=True)
        if self.dblp_url.endswith("gz"):
            with open(self.dblp_xml, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
        else:
            with gzip.open(self.dblp_xml, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
        print(f"DBLP database downloaded to {self.dblp_xml}.")

    def build(self, refresh=False, d=2, fields=None):
        """
        Main class method. Create the data and index files.

        Parameters
        ----------
        refresh: bool
            Tell if files are to be rebuilt if they are already there.
        d: int
            depth level where articles are. Usually 2 or 3 (2 for the main database).
        fields: set, optional
            Set of fields to collect. Default to :obj:`~gismo.datasets.dblp.DEFAULT_FIELDS`.

        Example
        -------
        By default, the class downloads the full dataset. Here we will limit to one entry.

        >>> toy_url = "https://dblp.org/pers/xx/m/Mathieu:Fabien.xml"
        >>> import tempfile
        >>> from gismo.filesource import FileSource
        >>> tmp = tempfile.TemporaryDirectory()
        >>> dblp = Dblp(dblp_url=toy_url, path=tmp.name)
        >>> dblp.build() # doctest.ELLIPSIS
        Retrieve https://dblp.org/pers/xx/m/Mathieu:Fabien.xml from the Internet.
        DBLP database downloaded to ...xml.gz.
        Converting DBLP database from ...xml.gz (may take a while).
        Building Index.
        Conversion done.

        By default, build uses existing files.

        >>> dblp.build() # doctest.ELLIPSIS
        File ...xml.gz already exists. Use refresh option to overwrite.
        File ...data already exists. Use refresh option to overwrite.

        The refresh parameter can be used to ignore existing files.

        >>> dblp.build(d=3, refresh=True) # doctest.ELLIPSIS
        Retrieve https://dblp.org/pers/xx/m/Mathieu:Fabien.xml from the Internet.
        DBLP database downloaded to ...xml.gz.
        Converting DBLP database from ...xml.gz (may take a while).
        Building Index.
        Conversion done.

        The resulting files can be used to create a FileSource.

        >>> source = FileSource(filename="dblp", path=tmp.name)
        >>> art = [s for s in source if s['title']=="Can P2P networks be super-scalable?"][0]
        >>> art['authors'] # doctest.ELLIPSIS
        ['François Baccelli', 'Fabien Mathieu', 'Ilkka Norros', 'Rémi Varloot']

        Don't forget to close source after use.

        >>> source.close()
        >>> tmp.cleanup()
        """
        if fields is None:
            fields = DEFAULT_FIELDS
        if self.dblp_xml.exists() and not refresh:
            print(f"File {self.dblp_xml} already exists. Use refresh option to overwrite.")
        else:
            print(f"Retrieve {self.dblp_url} from the Internet.")
            self.download()
        if self.dblp_data.exists() and not refresh:
            print(f"File {self.dblp_data} already exists. Use refresh option to overwrite.")
        else:
            print(f"Converting DBLP database from {self.dblp_xml} (may take a while).")
            # Download the DTD parser
            r = requests.get("https://dblp.uni-trier.de/xml/dblp.dtd")
            with open(self.path / Path("dblp.dtd"), 'w') as f:
                f.write(r.text)

            with gzip.open(self.dblp_xml, "rb") as f:
                index = [0]
                with open(self.dblp_data, "wb") as g:
                    context = etree.iterparse(f, events=('start', 'end',), load_dtd=True)
                    fast_iter(context, element_to_filesource, d=d, data_handler=g, index=index, fields=fields)
                print(f"Building Index.")
                with open(self.dblp_index, "wb") as g:
                    pickle.dump(np.array(index), g)
                print(f"Conversion done.")



import numpy as np
import logging
from scipy.sparse import csr_matrix, vstack, hstack

from gismo.common import auto_k, toy_source_text
from gismo.parameters import Parameters, DEFAULT_LANDMARKS_PARAMETERS
from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.gismo import Gismo
from gismo.clustering import subspace_clusterize, Cluster, subspace_distortion
from gismo.post_processing import post_landmarks_item_raw, post_landmarks_cluster_json

logging.basicConfig()
log = logging.getLogger("Gismo")


def get_direction(reference, balance):
    """
    Converts a reference object into a `n+m` direction (dense or sparse depending on reference type).

    Parameters
    ----------
    reference: Gismo or Landmarks or Cluster or np.ndarray or csr_matrix.
        The object from which a direction will be extracted.
    balance: float in range [0.0, 1.0]
        The trade-off between documents and features.
        Set to 0.0, only the feature space will be used.
        Set to 1.0, only the document space will be used.

    Returns
    -------
    np.ndarray or csr_matrix
        A `n+m` direction.
    """
    if isinstance(reference, Gismo):
        return np.hstack([balance * reference.diteration.x_relevance,
                          (1 - balance) * reference.diteration.y_relevance])
    if isinstance(reference, Landmarks):
        return np.hstack([balance * reference.x_direction, (1 - balance) * reference.y_direction])
    if isinstance(reference, Cluster):
        return reference.vector
    return reference


class Landmarks(Corpus):
    """
    The `Landmarks` class is a subclass :py:class:`~gismo.corpus.Corpus`.
    It offers the capability to batch-rank all its entries against a :py:class:`~gismo.gismo.Gismo` instance.
    After it has been processed, a `Landmarks` can be used to analyze/classify
    :py:class:`~gismo.gismo.Gismo` queries, :py:class:`~gismo.clustering.Cluster`,
    or :py:class:`~gismo.landmarks.Landmarks`.

    Landmarks also offers the possibility to reduce a source or a gismo to its neighborhood.
    This can be useful if the source is huge and one wants something smaller for performance.

    Parameters
    ------------

    source: list
        The list of items that form Landmarks.
    to_text: function
        The function that transforms an item into text
    kwargs: dict
        Custom default runtime parameters.
        You just need to specify the parameters that differ from :obj:`~gismo.parameters.DEFAULT_LANDMARKS_PARAMETERS`.

    Examples
    --------

    Landmarks lean on a Gismo. We can use a toy Gismo to start with.

    >>> corpus = Corpus(toy_source_text)
    >>> embedding = Embedding()
    >>> embedding.fit_transform(corpus)
    >>> gismo = Gismo(corpus, embedding)
    >>> print(toy_source_text) # doctest: +NORMALIZE_WHITESPACE
    ['Gizmo is a Mogwaï.',
    'This is a sentence about Blade.',
    'This is another sentence about Shadoks.',
    'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.',
    'In chinese folklore, a Mogwaï is a demon.']

    Landmarks are constructed exactly like a Gismo object, with a source and a `to_text` function.

    >>> landmarks_source = [{'name': 'Movies', 'content': 'Star Wars, Gremlins, and Blade are movies.'},
    ... {'name': 'Gremlins', 'content': 'The Gremlins movie features a Mogwai.'},
    ... {'name': 'Star Wars', 'content': 'The Star Wars movies feature Yoda.'},
    ... {'name': 'Shadoks', 'content': 'Shadoks is a French sarcastic show.'},]
    >>> landmarks = Landmarks(landmarks_source, to_text=lambda e: e['content'])

    The :func:`~gismo.landmarks.Landmarks.fit` method compute gismo queries for all landmarks and retain the results.

    >>> landmarks.fit(gismo)

    We run the request *Yoda* and look at the key landmarks.
    Note that *Gremlins* comes before *Star Wars*. This is actually *correct* in this small dataset:
    the word `Yoda` only exists in one sentence, which contains the words `Gremlins` and `Gizmo`.

    >>> success = gismo.rank('yoda')
    >>> landmarks.get_landmarks_by_rank(gismo) # doctest: +NORMALIZE_WHITESPACE
    [{'name': 'Gremlins', 'content': 'The Gremlins movie features a Mogwai.'},
    {'name': 'Star Wars', 'content': 'The Star Wars movies feature Yoda.'},
    {'name': 'Movies', 'content': 'Star Wars, Gremlins, and Blade are movies.'}]

    For better readibility, we set the item post_processing to return the `name` of a landmark item.

    >>> landmarks.post_item = lambda lmk, i: lmk[i]['name']
    >>> landmarks.get_landmarks_by_rank(gismo)
    ['Gremlins', 'Star Wars', 'Movies']

    The balance adjusts between documents and features spaces.
    A balance set to 1.0 focuses only on documents.

    >>> success = gismo.rank('blade')
    >>> landmarks.get_landmarks_by_rank(gismo, balance=1)
    ['Movies']

    A balance set to 0.0 focuses only on features.
    For *blade*, this triggers *Shadoks* as a secondary result, because of the shared word *sentence*.

    >>> landmarks.get_landmarks_by_rank(gismo, balance=0)
    ['Movies', 'Shadoks']

    Landmarks can be used to analyze landmarks.

    >>> landmarks.get_landmarks_by_rank(landmarks)
    ['Gremlins', 'Star Wars']

    See again how balance can change things.
    Here a balance set to 0.0 (using only features) fully changes the results.

    >>> landmarks.get_landmarks_by_rank(landmarks, balance=0)
    ['Shadoks']

    Like for :py:class:`~gismo.gismo.Gismo`, landmarks can provide clusters.

    >>> success = gismo.rank('gizmo')
    >>> landmarks.get_landmarks_by_cluster(gismo) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    {'landmark': {'name': 'Gremlins',
                  'content': 'The Gremlins movie features a Mogwai.'},
                  'focus': np.float64(0.999998...),
                  'children': [{'landmark': {'name': 'Gremlins',
                  'content': 'The Gremlins movie features a Mogwai.'},
                  'focus': 1.0, 'children': []},
                  {'landmark': {'name': 'Star Wars', 'content': 'The Star Wars movies feature Yoda.'},
                  'focus': 1.0, 'children': []},
                  {'landmark': {'name': 'Movies', 'content': 'Star Wars, Gremlins, and Blade are movies.'},
                  'focus': 1.0, 'children': []}]}

    We can set the `post_cluster` attribute to customize the output. Gismo provides a simple display.

    >>> from gismo.post_processing import post_landmarks_cluster_print
    >>> landmarks.post_cluster = post_landmarks_cluster_print
    >>> landmarks.get_landmarks_by_cluster(gismo) # doctest: +NORMALIZE_WHITESPACE
    F: 1.00.
    - Gremlins
    - Star Wars
    - Movies

    Like for :py:class:`~gismo.gismo.Gismo`, parameters like `k`, `distortion`, or `resolution` can be used.

    >>> landmarks.get_landmarks_by_cluster(gismo, k=4, distortion=False, resolution=.9) # doctest: +NORMALIZE_WHITESPACE
    F: 0.03.
    - F: 0.93.
    -- F: 1.00.
    --- Gremlins
    --- Star Wars
    -- Movies
    - Shadoks

    Note that a :py:class:`~gismo.clustering.Cluster` can also be used as reference for the
    :meth:`~gismo.landmarks.Landmarks.get_landmarks_by_rank` and
    :meth:`~gismo.landmarks.Landmarks.get_landmarks_by_cluster` methods.

    >>> cluster = landmarks.get_landmarks_by_cluster(gismo, post=False)
    >>> landmarks.get_landmarks_by_rank(cluster)
    ['Gremlins', 'Star Wars', 'Movies']

    Yet, you cannot use anything as reference. For example, you cannot use a string as such.

    >>> landmarks.get_landmarks_by_rank("Landmarks do not use external queries (pass them to a gismo")  # doctest.ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: bad operand type for unary -: 'NoneType'

    Last but not least, landmarks can be used to reduce the size of a source or a :py:class:`~gismo.gismo.Gismo`.
    The reduction is controlled by the `x_density` attribute that tells the number of documents each landmark will
    allow to keep.

    >>> landmarks.parameters.x_density = 1
    >>> reduced_gismo = landmarks.get_reduced_gismo(gismo)
    >>> reduced_gismo.corpus.source # doctest: +NORMALIZE_WHITESPACE
    ['This is another sentence about Shadoks.',
    'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the
    Gremlins movie by comparing Gizmo and Yoda.']

    Side remark #1: in the constructor, `to_text` indicates how to convert an item to `str`, while `ranking_function`
    specifies how to run a query on a :py:class:`~gismo.gismo.Gismo`. Yet, it is possible to have the text conversion
    handled by the `ranking_function`.

    >>> landmarks = Landmarks(landmarks_source, rank=lambda g, q: g.rank(q['content']))
    >>> landmarks.fit(gismo)
    >>> success = gismo.rank('yoda')
    >>> landmarks.post_item = lambda lmk, i: lmk[i]['name']
    >>> landmarks.get_landmarks_by_rank(gismo)
    ['Star Wars', 'Movies', 'Gremlins']

    However, this is bad practice. When you only need to customize the way an item is converted to text, you should
    stick to `to_text`. `ranking_function` is for more elaborated filters that require to change the default way
    gismo does queries.

    Side remark #2: if a landmark item query fails (its text does not intersect the gismo features),
    the default uniform projection will be used and a warning will be issued. This may yield to
    undesired results.

    >>> landmarks_source.append({'name': 'unrelated', 'content': 'unrelated.'})
    >>> landmarks = Landmarks(landmarks_source, to_text=lambda e: e['content'])
    >>> landmarks.fit(gismo)
    >>> success = gismo.rank('gizmo')
    >>> landmarks.post_item = lambda lmk, i: lmk[i]['name']
    >>> landmarks.get_landmarks_by_rank(gismo)
    ['Shadoks', 'unrelated']
    """

    def __init__(self, source=None, to_text=None, **kwargs):
        self.parameters = Parameters(parameter_list=DEFAULT_LANDMARKS_PARAMETERS, **kwargs)

        self.x_vectors = None
        self.y_vectors = None

        self.x_direction = None
        self.y_direction = None

        self.post_item = post_landmarks_item_raw
        self.post_cluster = post_landmarks_cluster_json

        super().__init__(source=source, to_text=to_text)

    def embed_entry(self, gismo, entry, **kwargs):
        p = self.parameters(**kwargs)
        log.debug(f"Processing {entry}.")
        success = p['rank'](gismo, entry)
        if not success:
            log.warning(f"Query {entry} didn't match any feature.")

        indptr = [0, min(p['y_density'], gismo.embedding.m)]
        indices = gismo.diteration.y_order[:p['y_density']]
        data = gismo.diteration.y_relevance[indices]
        y = csr_matrix((data, indices, indptr), shape=(1, gismo.embedding.m))

        indptr = [0, min(p['x_density'], gismo.embedding.n)]
        indices = gismo.diteration.x_order[:p['x_density']]
        data = gismo.diteration.x_relevance[indices]
        x = csr_matrix((data, indices, indptr), shape=(1, gismo.embedding.n))
        log.debug(f"Landmarks of {entry} computed.")
        return x / np.sum(x), y / np.sum(y)

    def fit(self, gismo, **kwargs):
        """
        Runs gismo queries on all landmarks.
        The relevance results are used to build two set of vectors:
        `x_vectors` is the vectors on the document space;
        `y_vectors` is the vectors on the document space.
        On each space, vectors are summed to build a direction, which is a sort of
        vector summary of the landmarks.

        Parameters
        ----------
        gismo: Gismo
            The Gismo on which vectors will be computed.
        kwargs: dict
            Custom Landmarks runtime parameters.

        Returns
        -------
        None

        """
        log.info(f"Start computation of {len(self)} landmarks.")
        xy = [self.embed_entry(gismo, entry, **kwargs) for entry in self.iterate_text()]
        self.x_vectors = vstack([v[0] for v in xy])
        self.x_direction = np.squeeze(np.asarray(np.sum(self.x_vectors, axis=0)))
        self.y_vectors = vstack([v[1] for v in xy])
        self.y_direction = np.squeeze(np.asarray(np.sum(self.y_vectors, axis=0)))
        log.info(f"All landmarks are built.")

    def get_base(self, balance):
        return csr_matrix(hstack([balance * self.x_vectors, (1 - balance) * self.y_vectors]))

    def get_landmarks_by_rank(self, reference, k=None, base=None, **kwargs):
        p = self.parameters(**kwargs)
        if base is None:
            base = self.get_base(p['balance'])
        direction = get_direction(reference, p['balance'])
        if isinstance(direction, np.ndarray):
            similarities = base.dot(direction)
        elif isinstance(direction, csr_matrix):
            similarities = np.squeeze(base.dot(direction.T).toarray())
        else:
            log.error("Direction type not supported. Direction must be gismo.gismo.Gismo, gismo.clustering.Cluster, "
                      "gismo.landmarks.Landmarks, numpy.ndarray or scipy.sparse.csr_matrix.")
            similarities = None
        order = np.argsort(-similarities)
        if k is None:
            k = auto_k(data=similarities, order=order, max_k=p['max_k'], target=p['target_k'])
        if p['post']:
            return [self.post_item(self, i) for i in order[:k]]
        else:
            return order[:k]

    def get_landmarks_by_cluster(self, reference, k=None, **kwargs):
        p = self.parameters(**kwargs)
        base = self.get_base(p['balance'])
        direction = get_direction(reference, p['balance'])
        order = self.get_landmarks_by_rank(reference=direction, k=k, base=base,
                                           target_k=p['target_k'], max_k=p['max_k'],
                                           balance=p['balance'], post=False)


        subspace = vstack([base[i, :] for i in order])
        if p['distortion']>0:
            subspace_distortion(indices=subspace.indices, data=subspace.data,
                                relevance=direction, distortion=p['distortion'])

        cluster = subspace_clusterize(subspace, resolution=p['resolution'], indices=order)
        if p['post']:
            return self.post_cluster(self, cluster)
        else:
            return cluster

    def get_reduced_source(self, gismo, rebuild=True):
        if rebuild:
            self.fit(gismo)
        reduced_indices = {i for i in self.x_vectors.indices}
        return [a for i, a in enumerate(gismo.corpus.source) if i in reduced_indices]

    def get_reduced_gismo(self, gismo, rebuild=True):
        reduced_corpus = Corpus(self.get_reduced_source(gismo, rebuild=rebuild),
                                to_text=gismo.corpus.to_text)
        reduced_embedding = Embedding(vectorizer=gismo.embedding.vectorizer)
        reduced_embedding.fit_transform(reduced_corpus)
        reduced_gismo = Gismo(reduced_corpus, reduced_embedding)
        reduced_gismo.parameters = gismo.parameters
        return reduced_gismo



from gismo.clustering import get_sim


def post_documents_item_raw(gismo, i):
    """
    Document indice to document entry

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        document indice

    Returns
    -------
    object
        Document i from corpus
    """
    return gismo.corpus[i]


def post_documents_item_content(gismo, i, max_size=None):
    """
    Document indice to document content.

    Assumes that document has a 'content' key.

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        document indice
    max_size: int, optional
        Maximum number of chars to return

    Returns
    -------
    str
        Content of document i from corpus
    """
    return gismo.corpus[i]['content'][:max_size]


def post_features_item_raw(gismo, i):
    """
    Feature indice to feature name

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        feature indice

    Returns
    -------
    str
        Feature i from embedding
    """
    return gismo.embedding.features[i]


def post_documents_cluster_json(gismo, cluster):
    """
    Convert cluster of documents into basic json

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of documents

    Returns
    -------
    dict
        dictionary with keys 'document', 'focus', and recursive 'children'
    """
    return {'document': gismo.corpus[cluster.indice],
            'focus': cluster.focus,
            'children': [post_documents_cluster_json(gismo, c) for c in cluster.children]}


def post_documents_cluster_print(gismo, cluster, post_item=None, depth=""):
    """
    Print an ASCII view of a document cluster with metrics (focus, relevance, similarity)

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of documents
    post_item: function, optional
        Post-processing function for individual documents
    depth: str, optional
        Current depth string used in recursion
    """
    if post_item is None:
        post_item = gismo.post_documents_item
    sim = get_sim(cluster.vector, gismo.diteration.y_relevance)
    if len(cluster.children) == 0:
        txt = post_item(gismo, cluster.indice)
        print(f"{depth} {txt} "
              f"(R: {gismo.diteration.x_relevance[cluster.indice]:.2f}; "
              f"S: {sim:.2f})")
    else:
        print(f"{depth} F: {cluster.focus:.2f}. "
              f"R: {sum(gismo.diteration.x_relevance[cluster.members]):.2f}. "
              f"S: {sim:.2f}.")
    for c in cluster.children:
        post_documents_cluster_print(gismo, c, post_item=post_item, depth=depth + '-')


def post_features_cluster_json(gismo, cluster):
    """
    Convert feature cluster into basic json

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of features

    Returns
    -------
    dict
        dictionary with keys 'feature', 'focus', and recursive 'children'
    """
    return {'feature': gismo.embedding.features[cluster.indice],
            'focus': cluster.focus,
            'children': [post_features_cluster_json(gismo, c) for c in cluster.children]}


def post_features_cluster_print(gismo, cluster, post_item=None, depth=""):
    """
        Print an ASCII view of a feature cluster with metrics (focus, relevance, similarity)

        Parameters
        ----------
        gismo: Gismo
            Gismo instance
        cluster: Cluster
            Cluster of features
        post_item: function, optional
            Post-processing function for individual features
        depth: str, optional
            Current depth string used in recursion
        """
    if post_item is None:
        post_item = gismo.post_features_item
    sim = get_sim(cluster.vector, gismo.diteration.x_relevance)
    if len(cluster.children) == 0:
        print(f"{depth} {post_item(gismo, cluster.indice)} "
              f"(R: {gismo.diteration.y_relevance[cluster.indice]:.2f}; "
              f"S: {sim:.2f})")
    else:
        print(f"{depth} F: {cluster.focus:.2f}. "
              f"R: {sum(gismo.diteration.y_relevance[cluster.members]):.2f}. "
              f"S: {sim:.2f}.")
    for c in cluster.children:
        post_features_cluster_print(gismo, c, depth=depth + '-')


def post_landmarks_item_raw(landmark, i):
    """
    Default post processor for individual landmarks.

    Parameters
    ----------
    landmark: Landmarks
        A Landmarks instance
    i: int
        Indice of the landmark to process.

    Returns
    -------
    object
        The landmark of indice i.
    """
    return landmark[i]


def post_landmarks_cluster_json(landmark, cluster):
    """
    Default post processor for a cluster of landmarks.

    Parameters
    ----------
    landmark: Landmarks
        A Landmarks instance
    cluster: Cluster
        Cluster of the landmarks to process.

    Returns
    -------
    dict
        A dict with the head landmark, cluster focus, and list of children.
    """
    return {'landmark': landmark[cluster.indice],
            'focus': cluster.focus,
            'children': [post_landmarks_cluster_json(landmark, child) for child in cluster.children]}


def post_landmarks_cluster_print(landmark, cluster, post_item=None, depth=""):
    """
    ASCII display post processor for a cluster of landmarks.

    Parameters
    ----------
    landmark: Landmarks
        A Landmarks instance
    cluster: Cluster
        Cluster of the landmarks to process.
    post_item: function, optional
        Post-processing function for individual landmarks
    depth: str, optional
        Current depth string used in recursion
    """
    if post_item is None:
        post_item = landmark.post_item
    if len(cluster.children) == 0:
        txt = post_item(landmark, cluster.indice)
        print(f"{depth} {txt} ")
    else:
        print(f"{depth} F: {cluster.focus:.2f}. ")
    for c in cluster.children:
        post_landmarks_cluster_print(landmark, c, post_item=post_item, depth=depth + '-')



import zlib
import json
import io
import dill as pickle
import numpy as np
from pathlib import Path
from gismo.corpus import toy_source_dict


def create_file_source(source=None, filename='mysource', path='.'):
    """
    Write a source (list of dict) to files in the same format used by FileSource. Only useful
    to transfer from a computer with a lot of RAM to a computer with less RAM. For more complex cases,
    e.g. when the initial source itself is a very large file, a dedicated converter has to be provided.

    Parameters
    ----------
    source: list of dict
        The source to write
    filename: str
        Stem of the file. Two files will be created, with suffixes *.index* and *.data*.
    path: str or Path
        Destination directory
    """
    if source is None:
        source = toy_source_dict
    path = Path(path)
    data_file = path / Path(f"{filename}.data")
    index_file = path / Path(f"{filename}.index")
    indices = [0]
    with open(data_file, "wb") as f:
        for item in source:
            f.write(zlib.compress(json.dumps(item).encode('utf8')))
            indices.append(f.tell())
    with open(index_file, "wb") as f:
        pickle.dump(indices, f)


class FileSource:
    """
    Yield a file source as a list. Assumes the existence of two files:
    The *mysource*.data file contains the stacked items. Each item is compressed with :py:mod:`zlib`;
    The  *mysource*.index files contains the list of pointers to seek items in the data file.

    The resulting source object is fully compatible with the :class:`~gismo.corpus.Corpus` class:
     * It can be iterated (``[item for item in source]``);
     * It can yield single items (``source[i]``);
     * It has a length (``len(source)``).

    More advanced functionalities like slices are not implemented.

    Parameters
    ----------
    path: str
                Location of the files
    filename: str
                Stem of the file
    load_source: bool
                Should the data be loaded in RAM

    Examples
    ---------
    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as dirname:
    ...    create_file_source(filename='mysource', path=dirname)
    ...    source = FileSource(filename='mysource', path=dirname, load_source=True)
    ...    content = [e['content'] for e in source]
    >>> content[:3]
    ['Gizmo is a Mogwaï.', 'This is a sentence about Blade.', 'This is another sentence about Shadoks.']

    Note: when source is read from file (``load_source=False``, default behavior), you need to close the source afterwards
    to avoid pending file handles.

    >>> with tempfile.TemporaryDirectory() as dirname:
    ...    create_file_source(filename='mysource', path=dirname)
    ...    source = FileSource(filename='mysource', path=dirname)
    ...    size = len(source)
    ...    item = source[0]
    ...    source.close()
    >>> size
    5
    >>> item
    {'title': 'First Document', 'content': 'Gizmo is a Mogwaï.'}
    """
    def __init__(self, filename="mysource", path='.', load_source=False):
        path = Path(path)
        index = path / Path(f"{filename}.index")
        data = path / Path(f"{filename}.data")
        # load index
        with open(index, "rb") as f:
            self.index = pickle.load(f)
        self.n = len(self.index) - 1
        if load_source:
            with open(data, "rb") as f:
                self.f = io.BytesIO(f.read())
        else:
            self.f = open(data, "rb")

    def __getitem__(self, i):
        self.f.seek(self.index[i])
        line = zlib.decompress(self.f.read(self.index[i + 1] - self.index[i])).decode('utf8')
        return json.loads(line)

    def __iter__(self):
        self.i = 0
        self.f.seek(0)
        return self

    def __next__(self):
        if self.i == self.n:
            raise StopIteration
        line = zlib.decompress(self.f.read(self.index[self.i + 1] - self.index[self.i])).decode('utf8')
        self.i += 1
        return json.loads(line)

    def __len__(self):
        return self.n

    def close(self):
        if self.f:
            self.f.close()
            self.f = None



#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own

from gismo.common import MixInIO, toy_source_text, toy_source_dict

import numpy as np
from itertools import chain


class Corpus(MixInIO):
    """
    The Corpus class is the starting point of any Gismo workflow. It abstracts dataset pre-processing.
    It is just a list of items (called documents in Gismo) augmented with a method that describes
    how to convert a document to a string object. It is used to build an :py:class:`~gismo.embedding.Embedding`.

    Parameters
    ----------
    source: list
        The list of items that constitutes the dataset to analyze. Actually, any iterable object with :func:`__len__`
        and :func:`__getitem__` methods can potentially be used as a source
        (see :py:class:`~gismo.filesource.FileSource` for an example).
    to_text: function, optional
        The function that transforms an item from the source into plain text (:py:class:`str`). If not set, it will
        default to the identity function ``lambda x: x``.

    Examples
    --------

    The following code uses the :py:obj:`~gismo.common.toy_source_text` list as source and specifies that the text
    extraction method should be: take the 15 first characters and add `...`.

    When we iterate with the :py:meth:`~gismo.corpus.Corpus.iterate` method, observe that the extraction is **not**
    applied.

    >>> corpus = Corpus(toy_source_text, to_text=lambda x: f"{x[:15]}...")
    >>> for c in corpus.iterate():
    ...    print(c)
    Gizmo is a Mogwaï.
    This is a sentence about Blade.
    This is another sentence about Shadoks.
    This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.
    In chinese folklore, a Mogwaï is a demon.

    When we iterate with the :py:meth:`~gismo.corpus.Corpus.iterate_text` method, observe that the extraction **is**
    applied.


    >>> for c in corpus.iterate_text():
    ...    print(c)
    Gizmo is a Mogw...
    This is a sente...
    This is another...
    This very long ...
    In chinese folk...

    A corpus object can be saved/loaded with the :py:meth:`~gismo.common.MixInIO.dump` and
    :py:meth:`~gismo.common.MixInIO.load` methods inherited from the MixIn :py:class:`~gismo.common.MixInIO` class.
    The :py:meth:`~gismo.common.MixInIO.load` method is a class method to be used instead of the constructor.

    >>> import tempfile
    >>> corpus1 = Corpus(toy_source_text)
    >>> with tempfile.TemporaryDirectory() as tmpdirname:
    ...    corpus1.dump(filename="myfile", path=tmpdirname)
    ...    corpus2 = Corpus.load(filename="myfile", path=tmpdirname)
    >>> corpus2[0]
    'Gizmo is a Mogwaï.'
    """
    def __init__(self, source=None, to_text=None):
        self.source = source
        self.i = 0
        self.n = 0 if source is None or not hasattr(source, '__len__') else len(source)
        self.iter = None
        if to_text is None:
            self.to_text = lambda x: x
        else:
            self.to_text = to_text

    def iterate_text(self, to_text=None):
        if to_text is None:
            to_text = self.to_text
        return (to_text(entry) for entry in self.source)

    def iterate(self):
        return (entry for entry in self.source)

    def __getitem__(self, i):
        return self.source[i]

    def __len__(self):
        return self.n

    def merge_new_source(self, new_source, doc2key=None):
        """
        Incorporate new entries while avoiding the creation of duplicates. This method is typically used when you have
        a dynamic source like a RSS feed and you want to periodically update your corpus.

        Parameters
        ----------
        new_source: list
                 Source compatible (e.g. similar item type) with the current source.
        doc2key: function
                 Callback that provides items with unique hashable keys, used to avoid duplicates.

        Examples
        --------

        The following code uses the :py:obj:`~gismo.common.toy_source_dict` list as source and add two new items,
        including a redundant one.

        >>> corpus = Corpus(toy_source_dict.copy(), to_text=lambda x: x['content'][:14])
        >>> len(corpus)
        5
        >>> new_corpus = [{"title": "Another document", "content": "I don't know what to say!"},
        ...     {'title': 'Fifth Document', 'content': 'In chinese folklore, a Mogwaï is a demon.'}]
        >>> corpus.merge_new_source(new_corpus, doc2key=lambda e: e['title'])
        >>> len(corpus)
        6
        >>> for c in corpus.iterate_text():
        ...    print(c)
        Gizmo is a Mog
        This is a sent
        This is anothe
        This very long
        In chinese fol
        I don't know w
        """
        if doc2key is None:
            print("Incremental corpus requires to provide a doc2key function")
            return self
        if self.source is None:
            self.source = []
        new_keys = {doc2key(d) for d in new_source} - {doc2key(d) for d in self.source}
        self.source += [d for d in new_source if doc2key(d) in new_keys]
        self.n = len(self.source)


class CorpusList(MixInIO):
    """
    This class makes a list of corpi behave like one single virtual corpus. This is useful to glue together corpi with
    distinct shapes and :py:meth:`to_text` methods.

    Parameters
    ----------
    corpus_list: list of :py:class:`.Corpus`
        The list of corpi to glue.


    Example
    -------
    >>> multi_corp = CorpusList([Corpus(toy_source_text, lambda x: x[:15]+"..."),
    ...                          Corpus(toy_source_dict, lambda e: e['title'])])
    >>> len(multi_corp)
    10
    >>> multi_corp[7]
    {'title': 'Third Document', 'content': 'This is another sentence about Shadoks.'}
    >>> for c in multi_corp.iterate_text():
    ...    print(c)
    Gizmo is a Mogw...
    This is a sente...
    This is another...
    This very long ...
    In chinese folk...
    First Document
    Second Document
    Third Document
    Fourth Document
    Fifth Document
    """

    def __init__(self, corpus_list=None, filename=None, path='.'):
        if corpus_list is None or len(corpus_list) == 0:
            print("Please provide a non-empty list of corpi!")
        else:
            self.corpus_list = corpus_list
            self.cum_n = np.cumsum([len(corpus) for corpus in self.corpus_list])
            self.n = self.cum_n[-1]

    def iterate(self):
        return chain.from_iterable([corpus.iterate() for corpus in self.corpus_list])

    def iterate_text(self):
        return chain.from_iterable([corpus.iterate_text() for corpus in self.corpus_list])

    def __getitem__(self, i):
        corpus_indice = np.searchsorted(self.cum_n, i, side='right')
        local_i = i if corpus_indice == 0 else (i - self.cum_n[corpus_indice - 1])
        return self.corpus_list[corpus_indice][local_i]

    def __len__(self):
        return self.n



"""Main module."""
import numpy as np
from scipy.sparse import vstack, csr_matrix
from sklearn.feature_extraction.text import CountVectorizer
from functools import partial

from gismo.common import MixInIO, toy_source_dict, auto_k
from gismo.datasets.dblp import url2source
from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.diteration import DIteration
from gismo.parameters import Parameters
from gismo.clustering import subspace_clusterize, covering_order, subspace_distortion
from gismo.post_processing import post_documents_item_raw, post_documents_item_content, post_documents_cluster_json, \
    post_features_item_raw, post_features_cluster_json, post_documents_cluster_print, post_features_cluster_print


class Gismo(MixInIO):
    """
    Gismo mixes a corpus and its embedding to provide search and structure methods.

    Parameters
    ----------
    corpus: Corpus
        Defines the documents of the gismo.
    embedding: Embedding
        Defines the embedding of the gismo.
    kwargs: dict
        Custom default runtime parameters.
        You just need to specify the parameters that differ from :obj:`~gismo.parameters.DEFAULT_PARAMETERS`.


    Example
    -------

    The Corpus class defines how documents of a source should be converted to plain text.

    >>> corpus = Corpus(toy_source_dict, lambda x: x['content'])

    The Embedding class extracts features (e.g. words) and computes weights between documents and features.

    >>> vectorizer = CountVectorizer(dtype=float)
    >>> embedding = Embedding(vectorizer=vectorizer)
    >>> embedding.fit_transform(corpus)
    >>> embedding.m # number of features
    36

    The Gismo class combines them for performing queries.
    After a query is performed, one can ask for the best items.
    The number of items to return can be specified with parameter ``k`` or automatically adjusted.

    >>> gismo = Gismo(corpus, embedding)
    >>> success = gismo.rank("Gizmo")
    >>> gismo.parameters.target_k = .2 # The toy dataset is very small, so we lower the auto_k parameter.
    >>> gismo.get_documents_by_rank()
    [{'title': 'First Document', 'content': 'Gizmo is a Mogwaï.'}, {'title': 'Fourth Document', 'content': 'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.'}, {'title': 'Fifth Document', 'content': 'In chinese folklore, a Mogwaï is a demon.'}]

    Post-processing functions can be used to tweak the returned object (the underlying ranking is unchanged)

    >>> gismo.post_documents_item = partial(post_documents_item_content, max_size=44)
    >>> gismo.get_documents_by_rank()
    ['Gizmo is a Mogwaï.', 'This very long sentence, with a lot of stuff', 'In chinese folklore, a Mogwaï is a demon.']

    Ranking also works on features.

    >>> gismo.get_features_by_rank()
    ['mogwaï', 'gizmo', 'is', 'chinese', 'demon', 'in', 'folklore']

    Clustering organizes results can provide additional hints on their relationships.

    >>> gismo.post_documents_cluster = post_documents_cluster_print
    >>> gismo.get_documents_by_cluster(resolution=.9) # doctest: +NORMALIZE_WHITESPACE
     F: 0.60. R: 0.65. S: 0.98.
    - F: 0.71. R: 0.57. S: 0.98.
    -- Gizmo is a Mogwaï. (R: 0.54; S: 0.99)
    -- In chinese folklore, a Mogwaï is a demon. (R: 0.04; S: 0.71)
    - This very long sentence, with a lot of stuff (R: 0.08; S: 0.69)
    >>> gismo.post_features_cluster = post_features_cluster_print
    >>> gismo.get_features_by_cluster() # doctest: +NORMALIZE_WHITESPACE
     F: 0.03. R: 0.29. S: 0.98.
    - F: 1.00. R: 0.27. S: 0.99.
    -- mogwaï (R: 0.12; S: 0.99)
    -- gizmo (R: 0.12; S: 0.99)
    -- is (R: 0.03; S: 0.99)
    - F: 1.00. R: 0.02. S: 0.07.
    -- chinese (R: 0.00; S: 0.07)
    -- demon (R: 0.00; S: 0.07)
    -- in (R: 0.00; S: 0.07)
    -- folklore (R: 0.00; S: 0.07)

    As an alternative to a textual query, the :meth:`~gismo.gismo.Gismo.rank` method can directly use a vector
    `z` as input.

    >>> z, s = gismo.embedding.query_projection("gizmo chinese folklore")
    >>> z # doctest: +NORMALIZE_WHITESPACE
     <Compressed Sparse Row sparse matrix of dtype 'float64'
        with 3 stored elements and shape (1, 36)>
    >>> s = gismo.rank(z=z)
    >>> s
    True
    >>> gismo.get_documents_by_rank(k=2)
    ['In chinese folklore, a Mogwaï is a demon.', 'Gizmo is a Mogwaï.']
    >>> gismo.get_features_by_rank()
    ['mogwaï', 'chinese', 'folklore', 'in', 'demon', 'gizmo', 'is']

    The class also offers :meth:`~gismo.gismo.Gismo.get_documents_by_coverage` and
    :meth:`~gismo.gismo.Gismo.get_features_by_coverage` that yield
    a list of results obtained from a Covering-like traversal of the ranked cluster.

    To demonstrate it, we first add an outsider document to the corpus and rebuild Gismo.

    >>> new_entry = {'title': 'Minority Report', 'content': 'Totally unrelated stuff.'}
    >>> corpus = Corpus(toy_source_dict+[new_entry], lambda x: x['content'])
    >>> vectorizer = CountVectorizer(dtype=float)
    >>> embedding = Embedding(vectorizer=vectorizer)
    >>> embedding.fit_transform(corpus)
    >>> gismo = Gismo(corpus, embedding)
    >>> gismo.post_documents_item = post_documents_item_content
    >>> success = gismo.rank("Gizmo")
    >>> gismo.parameters.target_k = .3

    Remind the classical rank-based result.

    >>> gismo.get_documents_by_rank()
    ['Gizmo is a Mogwaï.', 'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.', 'In chinese folklore, a Mogwaï is a demon.']

    Gismo can use the cluster to propose alternate results that try to cover more subjects.

    >>> gismo.get_documents_by_coverage()
    ['Gizmo is a Mogwaï.', 'Totally unrelated stuff.', 'This is a sentence about Blade.']

    Note how the new entry, which has nothing to do with the rest, is pushed into the results.
    By setting the ``wide`` option to False, we get an alternative that focuses on mainstream results.

    >>> gismo.get_documents_by_coverage(wide=False)
    ['Gizmo is a Mogwaï.', 'This is a sentence about Blade.', 'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.']

    The same principle applies for features.

    >>> gismo.get_features_by_rank()
    ['mogwaï', 'gizmo', 'is', 'chinese', 'demon', 'in', 'folklore']

    >>> gismo.get_features_by_coverage()
    ['mogwaï', 'about', 'chinese', 'and', 'gizmo', 'is', 'demon']
    """

    def __init__(self, corpus=None, embedding=None, **kwargs):
        self.corpus = corpus
        self.embedding = embedding
        self.diteration = DIteration(n=embedding.n, m=embedding.m)

        self.parameters = Parameters(**kwargs)

        self.post_documents_item = post_documents_item_raw
        self.post_features_item = post_features_item_raw
        self.post_documents_cluster = post_documents_cluster_json
        self.post_features_cluster = post_features_cluster_json

    # Ranking Part
    def rank(self, query="", z=None, **kwargs):
        """
        Runs the Diteration using query as starting point

        Parameters
        ----------
        query: str
            Text that starts DIteration
        z: :class:`~scipy.sparse.csr_matrix`, optional
            Query vector to use in place of the textual query
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        success: bool
            success of the query projection. If projection fails, a ranking on uniform distribution is performed.
        """
        p = self.parameters(**kwargs)
        if z is None:
            z, success = self.embedding.query_projection(query)
        else:
            success = True
        self.diteration(self.embedding.x, self.embedding.y, z,
                        alpha=p['alpha'], n_iter=p['n_iter'],
                        offset=p['offset'], memory=p['memory'])
        return success

    def get_documents_by_rank(self, k=None, **kwargs):
        """
        Returns a list of top documents according to the current ranking.
        By default, the documents are post_processed through the post_documents_item method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.


        Returns
        -------
        list
        """
        p = self.parameters(**kwargs)
        if k is None:
            k = auto_k(data=self.diteration.x_relevance,
                       order=self.diteration.x_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        if p['post']:
            return [self.post_documents_item(self, i) for i in self.diteration.x_order[:k]]
        else:
            return self.diteration.x_order[:k]

    def get_features_by_rank(self, k=None, **kwargs):
        """
        Returns a list of top features according to the current ranking.
        By default, the features are post_processed through the post_features_item method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        list
        """
        p = self.parameters(**kwargs)
        if k is None:
            k = auto_k(data=self.diteration.y_relevance,
                       order=self.diteration.y_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        if p['post']:
            return [self.post_features_item(self, i) for i in self.diteration.y_order[:k]]
        else:
            return self.diteration.y_order[:k]

    # Cluster part
    def get_documents_by_cluster_from_indices(self, indices, **kwargs):
        """
        Returns a cluster of documents.
        The cluster is by default post_processed through the post_documents_cluster method.

        Parameters
        ----------
        indices: list of int
            The indices of documents to be processed. It is assumed that the documents
            are sorted by importance.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        object

        """
        p = self.parameters(**kwargs)
        subspace = vstack([self.embedding.x[i, :] for i in indices])
        if p['distortion']>0:
            subspace_distortion(indices=subspace.indices, data=subspace.data,
                                relevance=self.diteration.y_relevance, distortion=p['distortion'])
        cluster = subspace_clusterize(subspace, p['resolution'], indices)
        if p['post']:
            return self.post_documents_cluster(self, cluster)
        return cluster

    def get_documents_by_cluster(self, k=None, **kwargs):
        """
        Returns a cluster of the best ranked documents.
        The cluster is by default post_processed through the post_documents_cluster method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        object

        """
        p = self.parameters(**kwargs)
        if k is None:
            k = auto_k(data=self.diteration.x_relevance,
                       order=self.diteration.x_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        return self.get_documents_by_cluster_from_indices(self.diteration.x_order[:k], **kwargs)

    def get_features_by_cluster_from_indices(self, indices, **kwargs):
        """
        Returns a cluster of features.
        The cluster is by default post_processed through the post_features_cluster method.

        Parameters
        ----------
        indices: list of int
            The indices of features to be processed. It is assumed that the features
            are sorted by importance.
        kwargs: dict, optional
            Custom runtime parameters

        Returns
        -------
        object

        """
        p = self.parameters(**kwargs)
        subspace = vstack([self.embedding.y[i, :] for i in indices])
        if p['distortion']>0:
            subspace_distortion(indices=subspace.indices, data=subspace.data,
                                relevance=self.diteration.x_relevance, distortion=p['distortion'])
        cluster = subspace_clusterize(subspace, p['resolution'], indices)
        if p['post']:
            return self.post_features_cluster(self, cluster)
        return cluster

    def get_features_by_cluster(self, k=None, **kwargs):
        """
        Returns a cluster of the best ranked features.
        The cluster is by default post_processed through the post_features_cluster method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        object
        """
        p = self.parameters(**kwargs)
        if k is None:
            k = auto_k(data=self.diteration.y_relevance,
                       order=self.diteration.y_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        return self.get_features_by_cluster_from_indices(self.diteration.y_order[:k], **kwargs)

    # Covering part

    def get_documents_by_coverage(self, k=None, **kwargs):
        """
        Returns a list of top covering documents.
        By default, the documents are post_processed through the post_documents_item method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        list

        """
        p = self.parameters(**kwargs)
        post = p['post']
        if k is None:
            k = auto_k(data=self.diteration.x_relevance,
                       order=self.diteration.x_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        p['post'] = False
        cluster = self.get_documents_by_cluster(k=int(k * p['stretch']),
                                                **p)
        indices = covering_order(cluster, wide=p['wide'])[:k]
        if post:
            return [self.post_documents_item(self, i) for i in indices]
        else:
            return indices

    def get_features_by_coverage(self, k=None, **kwargs):
        """
        Returns a list of top covering features.
        By default, the features are post_processed through the post_features_item method.

        Parameters
        ----------
        k: int, optional
            Number of documents to output. If not set, k is automatically computed
            using the max_k and target_k runtime parameters.
        kwargs: dict, optional
            Custom runtime parameters.

        Returns
        -------
        list

        """
        p = self.parameters(**kwargs)
        post = p['post']
        if k is None:
            k = auto_k(data=self.diteration.y_relevance,
                       order=self.diteration.y_order,
                       max_k=p['max_k'],
                       target=p['target_k'])
        p['post'] = False
        cluster = self.get_features_by_cluster(k=int(k * p['stretch']),
                                               **p)
        indices = covering_order(cluster, wide=p['wide'])[:k]
        if post:
            return [self.post_features_item(self, i) for i in indices]
        else:
            return indices


class XGismo(Gismo):
    """
    Given two distinct embeddings base on the same set of documents, builds a new gismo.
    The features of ``x_embedding`` are the corpus of this new gismo.
    The features of ``y_embedding`` are the features of this new gismo.
    The dual embedding of the new gismo is obtained by crossing the two input dual embeddings.

    xgismo behaves essentially as a gismo object. The main difference is an additional parameter ``y`` for the
    rank method, to control if the query projection should be performed on the ``y_embedding`` or on the
    ``x_embedding``.

    Parameters
    ----------
    x_embedding: Embedding
        The *left* embedding, which defines the documents of the xgismo.
    y_embedding: Embedding
        The *right* embedding, which defines the features of the xgismo.
    filename: str, optional
                If set, will load xgismo from file.
    path: str or Path, optional
        Directory where the xgismo is to be loaded from.
    kwargs: dict
        Custom default runtime parameters.
        You just need to specify the parameters that differ from :class:`~gismo.parameters.DEFAULT_PARAMETERS`.

    Examples
    ---------
    One the main use case for XGismo consists in transforming a list of articles into a Gismo that relates authors
    and the words they use. Let's start by retrieving a few articles.

    >>> toy_url = "https://dblp.org/pers/xx/m/Mathieu:Fabien.xml"
    >>> source = [a for a in url2source(toy_url) if int(a['year'])<2023]

    Then we build the embedding of words.

    >>> corpus = Corpus(source, to_text=lambda x: x['title'])
    >>> w_count = CountVectorizer(dtype=float, stop_words='english')
    >>> w_embedding = Embedding(w_count)
    >>> w_embedding.fit_transform(corpus)

    And the embedding of authors.

    >>> to_authors_text = lambda dic: " ".join([a.replace(' ', '_') for a in dic['authors']])
    >>> corpus.to_text = to_authors_text
    >>> a_count = CountVectorizer(dtype=float, preprocessor=lambda x:x, tokenizer=lambda x: x.split(' '))
    >>> a_embedding = Embedding(a_count)
    >>> a_embedding.fit_transform(corpus)

    We can now combine the two embeddings in one xgismo.

    >>> xgismo = XGismo(a_embedding, w_embedding)
    >>> xgismo.post_documents_item = lambda g, i: g.corpus[i].replace('_', ' ')

    We can use xgismo to query keyword(s).

    >>> success = xgismo.rank("Pagerank")
    >>> xgismo.get_documents_by_rank()
    ['Mohamed Bouklit', 'Dohy Hong', 'The Dang Huynh']

    We can use it to query researcher(s).

    >>> success = xgismo.rank("Anne_Bouillard", y=False)
    >>> xgismo.get_documents_by_rank()
    ['Anne Bouillard', 'Elie de Panafieu', 'Céline Comte', 'Philippe Sehier', 'Thomas Deiß', 'Dmitry Lebedev']
    """
    def __init__(self, x_embedding=None, y_embedding=None, filename=None, path=".", **kwargs):
        if filename is not None:
            self.load(filename=filename, path=path)
        else:
            embedding = Embedding()
            embedding.n = x_embedding.m
            embedding.m = y_embedding.m
            embedding.features = y_embedding.features
            embedding.x = np.dot(x_embedding.y, y_embedding.x)
            embedding.x_norm = np.ones(embedding.n)
            embedding.y = np.dot(y_embedding.y, x_embedding.x)
            embedding.y_norm = np.ones(embedding.m)
            embedding.idf = y_embedding.idf
            super().__init__(corpus=Corpus(x_embedding.features, to_text=lambda x: x), embedding=embedding, **kwargs)

            self.x_projection = x_embedding.query_projection
            self.y_projection = y_embedding.query_projection

    def rank(self, query="", y=True, **kwargs):
        """
        Runs the DIteration using query as starting point.
        ``query`` can be evaluated on features (``y=True``) or documents (``y=False``).

        Parameters
        ----------
        query: str
           Text that starts DIteration
        y: bool
           Determines if query should be evaluated on features (``True``) or documents (``False``).
        kwargs: dict, optional
            Custom runtime parameters.


        Returns
        -------
        success: bool
            success of the query projection. If projection fails, a ranking on uniform distribution is performed.
        """
        p = self.parameters(**kwargs)
        if y:
            z, found = self.y_projection(query)
            offset = 1.0
        else:
            z, found = self.x_projection(query)
            z = np.dot(z, self.embedding.x)
            offset = 0.0
        self.embedding._result_found = found
        self.diteration(self.embedding.x, self.embedding.y, z,
                        alpha=p['alpha'], n_iter=p['n_iter'],
                        offset=offset, memory=p['memory'])
        return found
